/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include "fitsio.h"
#include "imaging_fdata.h"
#include "delay.h"
#include "flag.h"
#define clight  2.9979250e8 /*m per sec */
#define  DELTAOPDPRISM 150.e-6
#define  DELTAOPDGRISM 800.e-6
#define  JUMPOPD  10.e-6
#define  SIDEDROP 1

void Usage() {
   printf("   oirAutoFlag inFile delayFile flagFile [-maxopd maxopd] [-minopd minopd] [-jump jumpopd] [-SN sigma\n");
   printf("      inFile = file name of output of oirRotateGroupDelay\n");
   printf("      delayFile  = file name of output of oirGroupDelay\n");
   printf("      flagFile   = file name of output file containing flag data\n");
   printf("      maxopd     = maximum allowed difference between tracking and true OPD (microns)\n");
   printf("        default  = 150 for PRISM and 800 for GRISM\n");
   printf("      minopd     = minimum allowed difference between tracking and true OPD (microns)\n");
   printf("         default = 0 \n");
   printf("      jumpopd    = maximum allowed jump in opd between frames (microns, default = 10\n");
   printf("      sigma      = minimum allowed signal/noise \n");
   return;
}
/********************************************************************
 * oirAutoFlag "flags" (i.e. sets a flag marking a point as invalid)
 * those points where the group delay differs from the tracked
 * value of the delay by more than the value of the parameter
 * "maxopd" or less than the value of the parameter "minopd".
 *
 * It also flags points where the group delay jumps more than
 * "jumpOpd" as measured by the absolute value of the 2nd 
 * difference of the group Delay, and then flags all points
 * within "sideDrop" points of such a jump
 ************************************************************************/


int main (int argc, char** argv) {

   int i, j, ierr;
   long nRow, iRow, sRow, oRow, pRow ;
   int nx;

   char outFileName[130];
   char *bang="!";

   fitsfile *inFile;
   fitsfile *outFile = NULL;
   int iKey ;
   double opd;
   imaging_fdata_table_def *idata;
   imaging_detector_table_def *DetectorDef;
   imaging_fdata_row *inData;

   delay_table_row *delayData;
   delay_table_def *dTable;
   double *delayTimes;
   double *groupDelays;
   float  *delayAmp;

   float maxOpd, minOpd, jumpOpd;
   float dOpd, jOpd;
   int sideDrop;
   long nJump, nSide, nMax, nGood, nMin, nAmp;

   char gris[12];

   flag_table_row *flagData;
   flag_table_def *flagTable;
   int *dFlags, *jFlags, *sFlags, *aFlags;

   double time0, time1;

   float dNoise;
   float snLevel;

   int iarg;

   ierr = 0;
   nJump  = 0;
   nSide  = 0;
   nMax   = 0;
   nMin   = 0;
   nAmp   = 0;
/*
      parse input; first set defaults then read command line
*/
   maxOpd = 0.;
   minOpd   = 0.;
   jumpOpd  = JUMPOPD;
   sideDrop = SIDEDROP;
   snLevel  = -1; /* no SN flagging by default */

   iarg = 4;
   while (iarg < argc ) {
      if (!strcmp("-jump", argv[iarg])) {
         sscanf(argv[++iarg],"%f",&jumpOpd);
         jumpOpd = jumpOpd * 1.e-6; /* microns to meters */
      }
      if (!strcmp("-maxopd", argv[iarg])) {
         sscanf(argv[++iarg],"%f",&maxOpd);
         maxOpd = maxOpd * 1.e-6; /* microns to meters */
      }
      if (!strcmp("-minopd", argv[iarg])) {
         sscanf(argv[++iarg],"%f",&minOpd);
         minOpd = minOpd * 1.e-6; /* microns to meters */
      }
      if (!strcmp("-SN", argv[iarg]))   sscanf(argv[++iarg],"%f",&snLevel);
      if (!strcmp("-side", argv[iarg])) sscanf(argv[++iarg],"%f",&sideDrop);
      ++iarg;
   }
/*
MAYBE sideDrop is useless
*/
sideDrop = 0;
   if (argc < 3) {
      Usage();
      return 1;
   }
   printf ("      ******   oirAutoFlag   ******\n");
   printf ("   Using dataFile %s and delFile %s \n", argv[1], argv[2]);
   printf ("   Using output File %s \n", argv[3]);

/*
      open delay file and read data into memory
*/
   fits_open_file (&inFile, argv[2], READONLY, &ierr);
   if (ierr != 0) {
      printf("ERROR:fits open delay file of %s gave error %i\n",argv[2],
         ierr);
      printf("%s \n",strerror(errno));
      return 1;
   }
   dTable = open_delay_table(inFile, &ierr);
   if (ierr != 0) {
      printf("ERROR:fits open delay table error %i\n", ierr);
      return 1;
   }
   i = fits_read_key(inFile, TFLOAT, "DNOISE", &dNoise, 0, &ierr);
   if (ierr != 0) {
      printf("   error reading dNoise from input delay file %i\n", ierr);
      printf("   ignoring signal/Noise flagging\n");
      dNoise = -1.;
      ierr   = 0;
   } else printf("   input delay amplitude noise is %f\n", dNoise);
   if ((snLevel <= 0) || (dNoise <= 0.))  printf ("   No SN flagging specified \n");
      else printf ("SN flagging specified at S/N = %f\n", snLevel);

   i = fits_get_num_rows (inFile,  &nRow, &ierr); 
   delayData = make_delay_table_row(dTable);
/*
         set up memory arrays
*/
   delayTimes  = (double *)malloc(nRow * sizeof(double));
   groupDelays = (double *)malloc(nRow * sizeof(double));
   delayAmp    = (float *)malloc(nRow * sizeof(float));
   dFlags      = (int *)malloc(nRow * sizeof(int));
   jFlags      = (int *)malloc(nRow * sizeof(int));
   sFlags      = (int *)malloc(nRow * sizeof(int));
   aFlags      = (int *)malloc(nRow * sizeof(int));


/*
         read delay data
*/
   for (iRow =1; iRow<=nRow; iRow++) {
      i = read_delay_table(inFile, dTable, delayData, iRow, &ierr);
      if (ierr != 0) {
         printf("ERROR:fits delay table read error %i row %li \n", ierr, iRow);
         return 1;
      }
   delayTimes[iRow-1]  = delayData->time;
   groupDelays[iRow-1] = clight * delayData->delay; /* meters */
   delayAmp[iRow-1]    = delayData->amplitude; 
   }
/*
      kill delay tables and close file;
*/
   kill_delay_table_def(dTable);
   kill_delay_table_row(delayData);
   fits_close_file(inFile, &ierr);


/*
        calculate jumps in groupOpd (from 2nd difference )and set flags if too large
*/
   for (iRow = 1; iRow <nRow-1; iRow++){
      jOpd = groupDelays[iRow] - 0.5 *(groupDelays[iRow-1] + groupDelays[iRow+1]);
      if (jOpd < 0.) jOpd = -jOpd;
      if (jOpd > jumpOpd) {
         jFlags[iRow] = 1;
      }
   }
   nJump = 0;
   for (iRow=1;iRow<nRow-1;iRow++) if (jFlags[iRow]) ++nJump;


/*
        use sideDrop to flag points next to jumps
*/
   for (iRow = 0; iRow < nRow; iRow++) {
      if (jFlags[iRow] != 0) for (sRow = -sideDrop; sRow <=sideDrop; sRow++) {
         if (sRow !=0) {  /* dont count flagged point again */
            pRow = iRow + sRow;
            if (pRow >= 0 && (pRow < nRow)) {
               sFlags[pRow] = 1;
            }
         }
      }
   }
   nSide = 0;
   for (iRow=0;iRow<nRow;iRow++) if (sFlags[iRow]) ++nSide;
/*
         amplitude clipping
*/
   nAmp = 0;
   if ((dNoise > 0.) && (snLevel > 0)) {
      snLevel = snLevel * dNoise;
      for (iRow=0;iRow<nRow;iRow++) if (delayAmp[iRow] < snLevel) {
         aFlags[iRow] = 1;
         ++nAmp;
      }
   }
/*
      open input file and get detector info
*/
   fits_open_file (&inFile, argv[1], READONLY, &ierr);
   if (ierr != 0) {
      printf("ERROR:fits openfile of %s gave error %i\n",argv[1],
         ierr);
      printf("%s \n",strerror(errno));
      return 1;
   }
/*    
      get grism setting and use to set opd diff limits
*/
   if (maxOpd == 0.) {
      i = fits_read_key(inFile, TSTRING, "HIERARCH ESO INS GRIS NAME", gris,
         0, &ierr);
      if (gris[0] == 'G') maxOpd = DELTAOPDGRISM;
         else maxOpd = DELTAOPDPRISM;
   }
   printf("   Using maxOpd = %5.1f mu, minOpd = %5.1f mu, jumpOpd = %5.1f mu, sideDrop = %i\n", 1.e6*maxOpd, 
      1.e6*minOpd, 1.e6*jumpOpd, sideDrop);
   DetectorDef = read_imaging_detector_table(inFile, &ierr);
   idata          = open_imaging_fdata_table(inFile, &ierr);

/*     
   get row size to be used as "NO_CHAN" in flag table
*/
   nx  = DetectorDef->naxis[0][0];
   inData = make_imaging_fdata_row(idata);

/* 
      now loop over input data, calculate OPD and compare to group delay
*/
   for (iRow=1; iRow<=nRow; iRow++) {
      i = read_imaging_fdata_table(inFile, idata, inData, iRow, &ierr);
      opd = inData->opd[1]+inData->local_opd[1] - inData->opd[0] - inData->local_opd[0];
      dOpd = opd - groupDelays[iRow-1];
      if (dOpd < 0.) dOpd = -dOpd;
      if (dOpd > maxOpd) {
         dFlags[iRow-1] = 1;
      }
      if (dOpd < minOpd) {
         dFlags[iRow-1] = 1;
      }
   }
   nMax = 0;
   for (iRow=1;iRow<nRow;iRow++) if(dFlags[iRow]) ++nMax;
/*
      shut down input data file
*/
   i = close_imaging_fdata_table(inFile, idata, &ierr);
   kill_imaging_fdata_row(inData);
/*
      create output flag table and row and fill in constant values
*/
   strcpy(outFileName, bang);
   strcat(outFileName, argv[3]);

   fits_create_file (&outFile, outFileName, &ierr);
   if (ierr !=0) {
      printf ("ERROR: opening fits flag output %s %i\n",argv[3], ierr);
      return 1;
      }
   flagTable = make_flag_table_def (outFile, "1776-07-04", nx, &ierr); 
   if (ierr !=0) {
      printf ("ERROR: %i creating flag data definition\n", ierr);
      return 1;
      }
   flagData = make_flag_table_row(flagTable);
   flagData->sourceId     =  1;
   flagData->telescope[0] =  1;
   flagData->telescope[1] = -1;
   flagData->chans[0]     =  1;
   flagData->chans[1]     = -1;

/*
         now compress maxOpd data and output
*/ 
   oRow = 1;
   iRow = 0;
   strcpy(flagData->reason, "GROUP - TRACKING DELAY");
   while (iRow < nRow) {
      if (dFlags[iRow] == 0) ++iRow; else {
         time0 = delayTimes[iRow];
         while(dFlags[iRow] != 0 && (iRow < nRow)) ++iRow;
         time1 = delayTimes[iRow-1];
         flagData->timeRange[0] = time0;
         flagData->timeRange[1] = time1;
         i = write_flag_table(outFile, flagTable, flagData, oRow++, &ierr);
         if (ierr != 0) {
            printf("ERROR:output error at line %li %i\n", oRow-1, ierr);
            return 1;
         }
      }
   }
/*
         now compress jumpOpd data (stored in jFlags) and output
         note there is nothing against flagging data twice
*/ 
   iRow = 0;
   strcpy(flagData->reason, "JUMP IN GROUP DELAY");
   while (iRow < nRow) {
      if (jFlags[iRow] == 0) ++iRow; else {
         time0 = delayTimes[iRow];
         while(jFlags[iRow] != 0 && (iRow < nRow)) ++iRow;
         time1 = delayTimes[iRow-1];
         flagData->timeRange[0] = time0;
         flagData->timeRange[1] = time1;
         i = write_flag_table(outFile, flagTable, flagData, oRow++, &ierr);
         if (ierr != 0) {
            printf("ERROR:output error at line %li %i\n", oRow-1, ierr);
            return 1;
         }
      }
   }
/*
         now compress amplitude data (stored in aFlags) and output
*/ 
   iRow = 0;
   strcpy(flagData->reason, "Low S/N");
   while (iRow < nRow) {
      if (aFlags[iRow] == 0) ++iRow; else {
         time0 = delayTimes[iRow];
         while(aFlags[iRow] != 0 && (iRow < nRow)) ++iRow;
         time1 = delayTimes[iRow-1];
         flagData->timeRange[0] = time0;
         flagData->timeRange[1] = time1;
         i = write_flag_table(outFile, flagTable, flagData, oRow++, &ierr);
         if (ierr != 0) {
            printf("ERROR:output error at line %li %i\n", oRow-1, ierr);
            return 1;
         }
      }
   }
/*
      cludge; flag last 100 rows
*/
  flagData->timeRange[0] = delayTimes[nRow-100];
  flagData->timeRange[1] = delayTimes[nRow-1];
  strcpy(flagData->reason, "END of OBS");
  i = write_flag_table(outFile, flagTable, flagData, oRow++, &ierr);
/*
      for my peace of mind, compute number of unflagged points
*/
   nGood  = 0;
   for (iRow=0;iRow<nRow;iRow++) 
      if ((dFlags[iRow]==0) && (jFlags[iRow]==0) && (sFlags[iRow]==0)
         && (aFlags[iRow]==0) ) nGood++;
/*
      shut down and shut up
*/
   kill_flag_table_def(flagTable);
   kill_flag_table_row(flagData);
   fits_close_file(outFile, &ierr);
   free(delayTimes);
   free(groupDelays);
   free(delayAmp);
   free(dFlags);
   free(jFlags);
   free(sFlags);
   free(aFlags);
   printf("   Out of %li input points\n", nRow);
   printf("   Flagged %li OPD jumps\n", nJump);
   printf("   Flagged %li points next to OPD jumps\n", nSide);
   printf("   Flagged %li points for excessive/insufficient OPD difference\n", nMax);
   printf("   Flagged %li points for insufficient S/N\n", nAmp);
/*   printf("   Flagged %li points for diminutive OPD difference\n", nMin); */
   printf("   There remain %li good points. \n", nGood);
   printf("   Note that some of the above flagged points may have been counted twice\n");
   printf("   e.g. an OPD jump may occur at an excessive OPD point.\n");
   return 0;
}
